/* This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
/** library to show BusVoodoo mode information on SSD1306 OLED display: name, activity, pinout (code)
 *  @file busvoodoo_oled.c
 *  @author King Kévin <kingkevin@cuvoodoo.info>
 *  @date 2018
 *  @note peripherals used: I2C @ref oled_ssd1306_i2c
 */
/* standard libraries */
#include <stdint.h> // standard integer types
#include <stdbool.h> // boolean type
#include <string.h> // string utilities

/* own libraries */
#include "global.h" // global utilities
#include "busvoodoo_oled.h" // own definitions
#include "oled_ssd1306.h" // OLED display utilities
#include "font.h" // font glyphs

/** if the OLED display is present and setup */
static bool busvoodoo_oled_present = false;

/** display pixel buffer */
static uint8_t busvoodoo_oled_display[128*8] = {0};

/** look-up table to swap the bit order in a byte
 *  @remark this is useful for the OLED screen since the top pixel is the MSb
 */
static const uint8_t bit_order_switch_lut[256] = { 0x00, 0x80, 0x40, 0xc0, 0x20, 0xa0, 0x60, 0xe0, 0x10, 0x90, 0x50, 0xd0, 0x30, 0xb0, 0x70, 0xf0, 0x08, 0x88, 0x48, 0xc8, 0x28, 0xa8, 0x68, 0xe8, 0x18, 0x98, 0x58, 0xd8, 0x38, 0xb8, 0x78, 0xf8, 0x04, 0x84, 0x44, 0xc4, 0x24, 0xa4, 0x64, 0xe4, 0x14, 0x94, 0x54, 0xd4, 0x34, 0xb4, 0x74, 0xf4, 0x0c, 0x8c, 0x4c, 0xcc, 0x2c, 0xac, 0x6c, 0xec, 0x1c, 0x9c, 0x5c, 0xdc, 0x3c, 0xbc, 0x7c, 0xfc, 0x02, 0x82, 0x42, 0xc2, 0x22, 0xa2, 0x62, 0xe2, 0x12, 0x92, 0x52, 0xd2, 0x32, 0xb2, 0x72, 0xf2, 0x0a, 0x8a, 0x4a, 0xca, 0x2a, 0xaa, 0x6a, 0xea, 0x1a, 0x9a, 0x5a, 0xda, 0x3a, 0xba, 0x7a, 0xfa, 0x06, 0x86, 0x46, 0xc6, 0x26, 0xa6, 0x66, 0xe6, 0x16, 0x96, 0x56, 0xd6, 0x36, 0xb6, 0x76, 0xf6, 0x0e, 0x8e, 0x4e, 0xce, 0x2e, 0xae, 0x6e, 0xee, 0x1e, 0x9e, 0x5e, 0xde, 0x3e, 0xbe, 0x7e, 0xfe, 0x01, 0x81, 0x41, 0xc1, 0x21, 0xa1, 0x61, 0xe1, 0x11, 0x91, 0x51, 0xd1, 0x31, 0xb1, 0x71, 0xf1, 0x09, 0x89, 0x49, 0xc9, 0x29, 0xa9, 0x69, 0xe9, 0x19, 0x99, 0x59, 0xd9, 0x39, 0xb9, 0x79, 0xf9, 0x05, 0x85, 0x45, 0xc5, 0x25, 0xa5, 0x65, 0xe5, 0x15, 0x95, 0x55, 0xd5, 0x35, 0xb5, 0x75, 0xf5, 0x0d, 0x8d, 0x4d, 0xcd, 0x2d, 0xad, 0x6d, 0xed, 0x1d, 0x9d, 0x5d, 0xdd, 0x3d, 0xbd, 0x7d, 0xfd, 0x03, 0x83, 0x43, 0xc3, 0x23, 0xa3, 0x63, 0xe3, 0x13, 0x93, 0x53, 0xd3, 0x33, 0xb3, 0x73, 0xf3, 0x0b, 0x8b, 0x4b, 0xcb, 0x2b, 0xab, 0x6b, 0xeb, 0x1b, 0x9b, 0x5b, 0xdb, 0x3b, 0xbb, 0x7b, 0xfb, 0x07, 0x87, 0x47, 0xc7, 0x27, 0xa7, 0x67, 0xe7, 0x17, 0x97, 0x57, 0xd7, 0x37, 0xb7, 0x77, 0xf7, 0x0f, 0x8f, 0x4f, 0xcf, 0x2f, 0xaf, 0x6f, 0xef, 0x1f, 0x9f, 0x5f, 0xdf, 0x3f, 0xbf, 0x7f, 0xff, };

void busvoodoo_oled_setup(void)
{

	// setup SSD1306 OLED display
	busvoodoo_oled_clear(); // clean display buffer
	busvoodoo_oled_present = oled_ssd1306_setup(0x3c); // setup OLED display
	if (busvoodoo_oled_present) {
#if DEBUG
		oled_ssd1306_test(); // test OLED display
#endif
		busvoodoo_oled_update(); // send display buffer
	};
}

void busvoodoo_oled_clear(void)
{
	// write all buffer to 0
	for (uint16_t i=0; i<LENGTH(busvoodoo_oled_display); i++) {
		busvoodoo_oled_display[i] = 0;
	}
}

#include "print.h"
void busvoodoo_oled_text_pos(uint8_t column, uint8_t row, enum font_name font_name, const char *text)
{
	// sanity checks
	if (column>=128) {
		return;
	}
	if (row>=64) {
		return;
	}
	if (font_name>=FONT_MAX) {
		return;
	}
	if (NULL==text) {
		return;
	}

	const struct font_s *font = &fonts[font_name]; // get selected font
	while (*text && column<128) {
		char c = *text;
		if (c>=' ' && c<' '+FONT_GLYPH_NUMBERS) {
			for (uint8_t i=0; i<font->width; i++) { // draw glyph from left to right
				uint8_t col = column+i; // calculate destination column position
				if (col>=128) {
					break; // end of screen reached
				}
				uint16_t glyph_column = font->glyphs[font->width*(c-' ')+i]; // get glyph column to draw
				// draw bottom part of glyph
				uint16_t pixel_byte_row = 128*((row/8)-0)+col;
				uint8_t glyph_byte_row = (glyph_column<<(7-(row%8)))>>0;
				glyph_byte_row = bit_order_switch_lut[glyph_byte_row];
				busvoodoo_oled_display[pixel_byte_row] |= glyph_byte_row;
				// draw middle part of glyph
				if (row>=8 && font->height>8-(row%8)) {
					pixel_byte_row -= 128;
					glyph_byte_row = (glyph_column<<(7-(row%8)))>>8;
					glyph_byte_row = bit_order_switch_lut[glyph_byte_row];
					busvoodoo_oled_display[pixel_byte_row] |= glyph_byte_row;
				}
				// draw top part of glyph
				if (row>=16 && font->height>8+(row%8)) {
					pixel_byte_row -= 128;
					glyph_byte_row = ((uint32_t)glyph_column<<(7-(row%8)))>>16;
					glyph_byte_row = bit_order_switch_lut[glyph_byte_row];
					busvoodoo_oled_display[pixel_byte_row] |= glyph_byte_row;
				}
			}
		}
		text++; // go to next character
		column += font->width+1;
	}
}

void busvoodoo_oled_text_left(char* text)
{
	// clear top (yellow) display part
	for (uint16_t i=0; i<128*2; i++) {
		 busvoodoo_oled_display[i] = 0;
	}

	// verify input
	if (NULL==text) {
		return;
	}

	busvoodoo_oled_text_pos(1, 14, FONT_KING14, text); // draw text on the left of top line
}

void busvoodoo_oled_text_right(char* text)
{
	// verify input
	if (NULL==text) {
		return;
	}

	// calculate column on which to start drawing
	uint8_t column = 0;
	if ((fonts[FONT_KING14].width+1)*strlen(text)<128) {
		column = 128-(fonts[FONT_KING14].width+1)*strlen(text);
	}
	busvoodoo_oled_text_pos(column, 14, FONT_KING14, text); // draw text on the right of the top line
}

void busvoodoo_oled_text_pinout(const char* pins[10], bool io_connector)
{
	// clear bottom (blue) display part
	for (uint16_t i=128*2; i<LENGTH(busvoodoo_oled_display); i++) {
		busvoodoo_oled_display[i] = 0;
	}

	// check input
	if (NULL==pins) {
		return;
	}

	// draw outline
	if (io_connector) {
		// top line
		for (uint8_t i=0; i<2+2+24*2; i++) {
			busvoodoo_oled_display[128*2+i] |= 0x03; // set left top two pixels
			busvoodoo_oled_display[128*2+2+2+24*3+i] |= 0x03; // set right top two pixels
		}
		// left and right lines
		for (uint8_t page=2; page<8; page++) {
			busvoodoo_oled_display[128*page+0] |= 0xff; // set left two pixels
			busvoodoo_oled_display[128*page+1] |= 0xff; // set left two pixels
			busvoodoo_oled_display[128*(page+1)-2] |= 0xff; // set right two pixels
			busvoodoo_oled_display[128*(page+1)-1] |= 0xff; // set right two pixels
		}
		// bottom line
		for (uint16_t i=128*7; i<128*8; i++) {
			busvoodoo_oled_display[i] |= 0xc0; // set bottom two pixels
		}
	} else {
		// middle line
		for (uint8_t i=0; i<2+2+24*2; i++) {
			busvoodoo_oled_display[128*4+i] |= 0x80; // set left bottom pixel
			busvoodoo_oled_display[128*4+2+2+24*3+i] |= 0x80; // set right bottom pixel
			busvoodoo_oled_display[128*5+i] |= 0x01; // set left top pixel
			busvoodoo_oled_display[128*5+2+2+24*3+i] |= 0x01; // set right top pixel
		}
		// left and right lines
		for (uint8_t page=5; page<8; page++) {
			busvoodoo_oled_display[128*page+0] |= 0xff; // set left two pixels
			busvoodoo_oled_display[128*page+1] |= 0xff; // set left two pixels
			busvoodoo_oled_display[128*(page+1)-2] |= 0xff; // set right two pixels
			busvoodoo_oled_display[128*(page+1)-1] |= 0xff; // set right two pixels
		}
	}

	for (uint8_t pin=0; pin<10; pin++) { // go through pin names
		if (NULL==pins[pin]) { // no text -> draw cross
			uint16_t column = 2+2+24*(4-pin/2)+2; // the start column to start drawing (from left)
			if (0==pin%2) {
				for (uint8_t col=0; col<20; col++) {
					uint32_t cross = (1<<col)|(1<<(19-col)); // the two dots of the cross
					if (io_connector) {
						busvoodoo_oled_display[128*2+column+col] |= cross<<3;
						busvoodoo_oled_display[128*3+column+col] |= cross>>5;
						busvoodoo_oled_display[128*4+column+col] |= cross>>13;
					} else {
						busvoodoo_oled_display[128*2+column+col] |= cross<<1;
						busvoodoo_oled_display[128*3+column+col] |= cross>>7;
						busvoodoo_oled_display[128*4+column+col] |= cross>>15;
					}
				}
			} else {
				for (uint8_t col=0; col<20; col++) {
					uint32_t cross = (1<<col)|(1<<(19-col)); // the two dots of the cross
					if (io_connector) {
						busvoodoo_oled_display[128*5+column+col] |= cross<<1;
						busvoodoo_oled_display[128*6+column+col] |= cross>>7;
						busvoodoo_oled_display[128*7+column+col] |= cross>>15;
					} else {
						busvoodoo_oled_display[128*5+column+col] |= cross<<3;
						busvoodoo_oled_display[128*6+column+col] |= cross>>5;
						busvoodoo_oled_display[128*7+column+col] |= cross>>13;
					}
				}
			}
		} else if (0==strlen(pins[pin])) {
			// leave blank
		} else if (strlen(pins[pin])<4) {
			// calculate start position on x-axis based on number of characters to put in 24x20 px box
			uint8_t column = 2+2+24*(4-pin/2)+(24-(fonts[FONT_KING10].width+1)*strlen(pins[pin])-1)/2+1;
			if (0==pin%2) {
				if (io_connector) {
					busvoodoo_oled_text_pos(column, 16+2+1+20/2+fonts[FONT_KING10].height/2, FONT_KING10, pins[pin]); // draw pin name
				} else {
					busvoodoo_oled_text_pos(column, 16+2+20/2+fonts[FONT_KING10].height/2, FONT_KING10, pins[pin]); // draw pin name
				}
			} else {
				if (io_connector) {
					busvoodoo_oled_text_pos(column, 16+2+1+20+1+1+20/2+fonts[FONT_KING10].height/2, FONT_KING10, pins[pin]); // draw pin name
				} else {
					busvoodoo_oled_text_pos(column, 16+2+20+1+2+1+20/2+fonts[FONT_KING10].height/2, FONT_KING10, pins[pin]); // draw pin name
				}
			}
		} else if (4==strlen(pins[pin])) {
			uint16_t column = 2+2+24*(4-pin/2); // start position on x-axis (for 4 characters on a 24 px box)
			if (0==pin%2) {
				if (io_connector) {
					busvoodoo_oled_text_pos(column, 16+2+1+20/2+fonts[FONT_KING8].height/2, FONT_KING8, pins[pin]); // draw pin name
				} else {
					busvoodoo_oled_text_pos(column, 16+2+20/2+fonts[FONT_KING8].height/2, FONT_KING8, pins[pin]); // draw pin name
				}
			} else {
				if (io_connector) {
					busvoodoo_oled_text_pos(column, 16+2+1+20+1+1+20/2+fonts[FONT_KING8].height/2, FONT_KING8, pins[pin]); // draw pin name
				} else {
					busvoodoo_oled_text_pos(column, 16+2+20+1+2+1+20/2+fonts[FONT_KING8].height/2, FONT_KING8, pins[pin]); // draw pin name
				}
			}
		} else {
			uint16_t column = 2+2+24*(4-pin/2); // start position on x-axis (for 4 characters on a 24 px box)
			char line_top[5] = {0, 0, 0, 0, 0};
			for (uint8_t c=0; c<4 && c<strlen(pins[pin]); c++) {
				line_top[c] = pins[pin][c];
			}
			char line_bottom[5] = {0, 0, 0, 0, 0};
			for (uint8_t c=4; c<8 && c<strlen(pins[pin]); c++) {
				line_bottom[c-4] = pins[pin][c];
			}
			if (0==pin%2) {
				if (io_connector) {
					busvoodoo_oled_text_pos(column, 16+2+1+20/4+fonts[FONT_KING8].height/2, FONT_KING8, line_top); // draw pin name
					busvoodoo_oled_text_pos(column, 16+2+1+20/2+20/4+fonts[FONT_KING8].height/2, FONT_KING8, line_bottom); // draw pin name
				} else {
					busvoodoo_oled_text_pos(column, 16+2+20/4+fonts[FONT_KING8].height/2, FONT_KING8, line_top); // draw pin name
					busvoodoo_oled_text_pos(column, 16+2+20/2+20/4+fonts[FONT_KING8].height/2, FONT_KING8, line_bottom); // draw pin name
				}
			} else {
				if (io_connector) {
					busvoodoo_oled_text_pos(column, 16+2+1+20+1+1+20/4+fonts[FONT_KING8].height/2, FONT_KING8, line_top); // draw pin name
					busvoodoo_oled_text_pos(column, 16+2+1+20+1+1+20/2+20/4+fonts[FONT_KING8].height/2, FONT_KING8, line_bottom); // draw pin name
				} else {
					busvoodoo_oled_text_pos(column, 16+2+20+1+2+1+20/4+fonts[FONT_KING8].height/2, FONT_KING8, line_top); // draw pin name
					busvoodoo_oled_text_pos(column, 16+2+20+1+2+1+1+20/2+20/4+fonts[FONT_KING8].height/2, FONT_KING8, line_bottom); // draw pin name
				}
			}
		}
	}
}

void busvoodoo_oled_update(void)
{
	if (busvoodoo_oled_present) { // only do something if the display is present
		oled_ssd1306_off(); // switch display off to not see the update
		oled_ssd1306_display(busvoodoo_oled_display, LENGTH(busvoodoo_oled_display)); // send current display buffer
		oled_ssd1306_on(); // switch display back on
	}
}
